import torch
import torch.nn as nn
import numpy as np
import task_generator as tg
import argparse
import random
import cv2
from torch.autograd import Variable

parser = argparse.ArgumentParser(description="One Shot Visual Recognition")
parser.add_argument("-c","--class_num",type = int, default = 5)
parser.add_argument("-s","--sample_num_per_class",type = int, default = 1)
parser.add_argument("-b","--batch_num_per_class",type = int, default = 19)
parser.add_argument("-u","--hidden_unit", type=int, default=10)
args = parser.parse_args()


# Hyper Parameters
CLASS_NUM = args.class_num
SAMPLE_NUM_PER_CLASS = 1
BATCH_NUM_PER_CLASS = args.batch_num_per_class

saved_models_filepath = './24层特征和三维相减/saved_models/'


class CNNEncoder(nn.Module):
    """docstring for ClassName"""
    def __init__(self):
        super(CNNEncoder, self).__init__()
        self.layer1 = nn.Sequential(
                        nn.Conv2d(3,64,kernel_size=3,padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer2 = nn.Sequential(
                        nn.Conv2d(64,64,kernel_size=3,padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer3 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())
        self.layer4 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())

    def forward(self, x):
        out1 = self.layer1(x)
        out2 = self.layer2(out1)
        out3 = self.layer3(out2)
        out4 = self.layer4(out3)
        return out1, out2, out3, out4


class ReshapeNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self):
        super(ReshapeNetwork, self).__init__()
        self.layer = nn.AvgPool2d(2)

    def forward(self, x):
        out = self.layer(x)
        return out


def main():
    metatrain_folders, metatest_folders = tg.mini_imagenet_folders()
    feature_encoder = CNNEncoder()
    reshapenetwork = ReshapeNetwork()
    feature_encoder.load_state_dict(torch.load(
        str("{}/miniimagenet_feature_encoder.pkl".format(saved_models_filepath))))
    print("load feature encoder success")

    degrees = random.choice([0, 90, 180, 270])
    task = tg.MiniImagenetTask(metatrain_folders, CLASS_NUM, SAMPLE_NUM_PER_CLASS, SAMPLE_NUM_PER_CLASS)
    sample_dataloader = tg.get_mini_imagenet_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                                         split="train", shuffle=False)
    batch_dataloader = tg.get_mini_imagenet_data_loader(task, num_per_class=BATCH_NUM_PER_CLASS,
                                                        split="test", shuffle=True)
    sample_images, sample_labels = sample_dataloader.__iter__().next()
    batches, batch_labels = batch_dataloader.__iter__().next()
    # sample_images.shape torch.Size([5, 1, 28,28])

    features1, features2, features3, features4 = feature_encoder(Variable(sample_images))
    batch_features1, batch_features2,batch_features3, batch_features4 = feature_encoder(Variable(batches))
    # features1.shape torch.Size([5, 64, 13, 13])
    # features2.shape torch.Size([5, 64, 5, 5])
    # features3.shape torch.Size([5, 64, 5, 5])
    # features4.shape torch.Size([5, 64, 5, 5])

    # print(sample_features1.shape) ([5, 64, 20, 20])
    # sample_features2 = reshapenetwork(features2)
    # print(sample_features2.shape) torch.Size([5, 64, 9, 9])
    # sample_features3 = reshapenetwork(features3)
    # print(sample_features3.shape) torch.Size([5, 64, 9, 9])
    # sample_features4 = reshapenetwork(features4)
    # print(sample_features4.shape) torch.Size([5, 64, 9, 9])
    multi_features24 = torch.cat((features2,features4),1)
    multi_features34 = torch.cat((features3,features4),1)
    # print(multi_features.shape)

    # 考虑到features的形状为：[batch_size, filter_nums, H, W] 提取其中的第一个过滤器得到的结果feature。
    # 以一张图片作为输入的情况下，我们得到的feature即为[H,W]大小的tensor。
    # 将tensor转为numpy，然后归一化到[0,1]，最后乘255，使得范围为[0,255]
    # 得到灰度图像并保存。
    multi_feature24 = multi_features24[1,30, :, :]
    multi_feature34 = multi_features34[1,30, :, :]
    # feature = batch_features4[1, 0, :, :]
    # batch = batches[1, 0, :, :]
    # print(sample_image.shape)
    # torch.Size([5, 13, 13])
    # feature = feature.data.numpy()
    # batch = batch.data.numpy()
    multi_feature24 = multi_feature24.data.numpy()
    multi_feature34 = multi_feature34.data.numpy()

    # use sigmod to [0,1]
    # feature = 1.0 / (1 + np.exp(-1 * feature))
    # batch = 1.0 / (1 + np.exp(-1 * batch))
    multi_feature24 = 1.0 / (1 + np.exp(-1 * multi_feature24))
    multi_feature34 = 1.0 / (1 + np.exp(-1 * multi_feature34))
    # to [0,255]
    # feature = np.round(feature * 255)
    # batch = np.round(batch * 255)
    multi_feature24 = 1.0 / (1 + np.exp(-1 * multi_feature24))
    multi_feature34 = 1.0 / (1 + np.exp(-1 * multi_feature34))
    # print(feature)
    # cv2.imwrite('./out4.jpg', feature)
    # cv2.imwrite('./batch.jpg', batch)
    cv2.imwrite('./multi_feature24.jpg', multi_feature24)
    cv2.imwrite('./multi_feature34.jpg', multi_feature34)


if __name__ == '__main__':
    main()
