'''三维相减后再拉直'''

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np
import task_generator_test as tg
import os
import math
import argparse
import scipy as sp
import scipy.stats
from storage import save_statistics, build_experiment_folder

parser = argparse.ArgumentParser(description="One Shot Visual Recognition")
parser.add_argument("-f","--feature_dim",type = int, default = 64)
parser.add_argument("-r","--relation_dim",type = int, default = 8)
parser.add_argument("-r3", "--relation3_dim", type=int, default=8)
parser.add_argument("-w","--class_num",type = int, default=5)
parser.add_argument("-s","--sample_num_per_class",type = int, default=1)
parser.add_argument("-b","--batch_num_per_class",type = int, default=15)
parser.add_argument("-e","--episode",type = int, default= 500000)
parser.add_argument("-t","--test_episode", type = int, default=600)
parser.add_argument("-l","--learning_rate", type = float, default=0.001)
parser.add_argument("-u","--hidden_unit",type=int,default=10)
parser.add_argument("-tl", "--experiment_title", type=str, default="experiment-title")
args = parser.parse_args()

# Hyper Parameters
FEATURE_DIM = args.feature_dim
RELATION_DIM = args.relation_dim
CLASS_NUM = args.class_num
SAMPLE_NUM_PER_CLASS = 1
BATCH_NUM_PER_CLASS = args.batch_num_per_class
EPISODE = args.episode
TEST_EPISODE = args.test_episode
LEARNING_RATE = args.learning_rate
HIDDEN_UNIT = args.hidden_unit
EXPREIMENT_TITLE = args.experiment_title

x = torch.device("cuda")


def mean_confidence_interval(data, confidence=0.95):
    a = 1.0*np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    h = se * sp.stats.t._ppf((1+confidence)/2., n-1)
    return m, h


class ReshapeNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self):
        super(ReshapeNetwork, self).__init__()
        self.layer = nn.AvgPool2d(2)

    def forward(self, x):
        out = self.layer(x)
        return out


class CNNEncoder(nn.Module):
    """docstring for ClassName"""
    def __init__(self):
        super(CNNEncoder, self).__init__()
        self.layer1 = nn.Sequential(
                        nn.Conv2d(3, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer2 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer3 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())
        self.layer4 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())

    def forward(self, x):
        out1 = self.layer1(x)
        # print('out1.size', out1.size())
        out2 = self.layer2(out1)
        # print('out2.size', out2.size())
        out3 = self.layer3(out2)
        # print('out3.size', out3.size())
        out4 = self.layer4(out3)
        # print('out4.size', out4.size())
        return out3, out4


class RelationNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self,input_size,hidden_size):
        super(RelationNetwork, self).__init__()
        self.layer1 = nn.Sequential(
                        nn.Conv2d(256,64,kernel_size=3,padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer2 = nn.Sequential(
                        nn.Conv2d(64,64,kernel_size=3,padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.fc1 = nn.Linear(input_size*3*3,hidden_size)
        self.fc2 = nn.Linear(hidden_size,1)

    def forward(self,x):
        out = self.layer1(x)
        out = self.layer2(out)
        out = out.view(out.size(0),-1)
        out = F.relu(self.fc1(out))
        out = F.sigmoid(self.fc2(out))
        return out


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
        m.weight.data.normal_(0, math.sqrt(2. / n))
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('BatchNorm') != -1:
        m.weight.data.fill_(1)
        m.bias.data.zero_()
    elif classname.find('Linear') != -1:
        n = m.weight.size(1)
        m.weight.data.normal_(0, 0.01)
        m.bias.data = torch.ones(m.bias.data.size())


def main():
    # Step 1: init data folders
    print("init data folders")
    # init character folders for dataset construction
    metatest_folders = tg.mini_imagenet_folders()

    # Step 2: init neural networks
    print("init neural networks")
    reshapenetwork = ReshapeNetwork()
    feature_encoder = CNNEncoder()
    relation_network = RelationNetwork(FEATURE_DIM,RELATION_DIM)

    reshapenetwork.to(x)
    feature_encoder.to(x)
    relation_network.to(x)

    saved_models_filepath, logs_filepath = build_experiment_folder(EXPREIMENT_TITLE)
    save_statistics(logs_filepath, ["episode", "test accuracy"], create=True)

    if os.path.exists(str("{}/miniimagenet_feature_encoder.pkl".format(saved_models_filepath))):
        feature_encoder.load_state_dict(torch.load(
            str("{}/miniimagenet_feature_encoder.pkl".format(saved_models_filepath))))
        print("load feature encoder success")
    if os.path.exists(str("/{}/miniimagenet_relation_network.pkl".format(saved_models_filepath))):
        relation_network.load_state_dict(torch.load(
            str("{}/miniimagenet_relation_network.pkl".format(saved_models_filepath))))
        print("load relation network success")

    # Step 3: build graph
    total_accuracy = 0.0
    for episode in range(EPISODE):
            # test
            print("Testing...")
            accuracies = []
            num_per_class = 3
            for i in range(TEST_EPISODE):
                total_rewards = 0
                counter = 0
                task = tg.MiniImagenetTask(metatest_folders, CLASS_NUM, 1, num_per_class)
                sample_dataloader = tg.get_mini_imagenet_data_loader(task,num_per_class=1,split="train",shuffle=False)
                test_dataloader = tg.get_mini_imagenet_data_loader(task,num_per_class=num_per_class,split="test",shuffle=True)
                sample_images,sample_labels = sample_dataloader.__iter__().next()
                for test_images,test_labels in test_dataloader:
                    test_labels = test_labels.long().to(x)
                    batch_size = test_labels.shape[0]
                    # calculate features
                    sample_features3, sample_features4 = feature_encoder(Variable(sample_images).to(x))
                    test_features3, test_features4 = feature_encoder(Variable(test_images).to(x))

                    # calculate relations
                    sample_features3 = reshapenetwork(sample_features3.to(x))
                    test_features3 = reshapenetwork(test_features3.to(x))
                    sample_features4 = reshapenetwork(sample_features4.to(x))
                    test_features4 = reshapenetwork(test_features4.to(x))

                    sample_features = torch.cat((sample_features3, sample_features4), 1)
                    # print("sample_features ", sample_features.size()) ([5, 128, 9, 9])
                    test_features = torch.cat((test_features3, test_features4), 1)
                    # print("test_features", test_features.size()) ([15, 128, 9, 9])

                    # calculate relations
                    sample_features_ext = sample_features.unsqueeze(0).repeat(batch_size, 1, 1, 1, 1)
                    test_features_ext = test_features.unsqueeze(0).repeat(1 * CLASS_NUM, 1, 1, 1, 1)
                    test_features_ext = torch.transpose(test_features_ext, 0, 1)
                    relation_pairs = torch.cat((sample_features_ext, test_features_ext), 2).view(-1, FEATURE_DIM * 2,
                                                                                                 9, 9)
                    relations = relation_network(relation_pairs).view(-1, CLASS_NUM)

                    _, predict_labels = torch.max(relations.data, 1)
                    rewards = [1 if predict_labels[j] == test_labels[j] else 0 for j in range(batch_size)]

                    total_rewards += np.sum(rewards)
                    counter += batch_size
                accuracy = total_rewards/1.0/counter
                accuracies.append(accuracy)

            test_accuracy, h = mean_confidence_interval(accuracies)

            print("test accuracy:", test_accuracy, "h:", h)
            save_statistics(logs_filepath, [episode, test_accuracy])
            total_accuracy += test_accuracy
    aver_accuracy = total_accuracy/EPISODE
    print("aver_accuracy:", aver_accuracy)
    save_statistics(logs_filepath, ["aver_accuracy"])
    save_statistics(logs_filepath, [ aver_accuracy])


if __name__ == '__main__':
    main()